function elementIsEnabled(element) {    if (document.doctype && document.doctype.name === 'xml')        return false;    let tagNamesToSearchForDisabledAttribute = new Set(['BUTTON', 'INPUT', 'SELECT', 'TEXTAREA']);    if (tagNamesToSearchForDisabledAttribute.has(element.tagName) && element.disabled)        return false;    let descendantOfDisabledFieldset = element.matches(`fieldset:disabled ${element.tagName}`);    let descendantOfFirstLegendInsideDisabledFieldset = element.matches(`fieldset:disabled > legend:first-child ${element.tagName}`);    let isDisabled = descendantOfDisabledFieldset && !descendantOfFirstLegendInsideDisabledFieldset;    return !isDisabled;}
// sourceURL=__InjectedScript_WebDriver_IsElementEnabled.js
